from pathlib import Path
from typing import Optional

import pandas as pd
from tqdm.auto import tqdm

from wpextract.extractors.data.links import LinkRegistry
from wpextract.extractors.io import load_df
from wpextract.parse.content import extract_content_data
from wpextract.parse.html import extract_html_text, parse_html
from wpextract.util.locale import extract_locale

EXPORT_COLUMNS = [
    "author",
    "content.rendered",
    "content.text",
    "date_gmt",
    "embeds",
    "excerpt.rendered",
    "excerpt.text",
    "featured_media",
    "images",
    "link",
    "link_locale",
    "links.external",
    "links.internal",
    "modified_gmt",
    "parent",
    "slug",
    "template",
    "title.rendered",
    "yoast_head_json.title",
]


RENAME_COLUMNS = {
    "content.rendered": "content.html",
    "title.rendered": "title.html",
    "excerpt.rendered": "excerpt.html",
    "yoast_head_json.title": "page_title",
}


def load_pages(path: Path, link_registry: LinkRegistry) -> Optional[pd.DataFrame]:
    """Load the pages from a JSON file.

    The JSON file is expected to be in the response format of the WordPress posts API.

    Args:
        path: The path to the JSON file
        link_registry: The link registry to populate

    Returns:
        A dataframe of the pages
    """
    pages_df = load_df(path)

    if pages_df is None:
        return None

    pages_df["date_gmt"] = pd.to_datetime(pages_df["date_gmt"])
    pages_df["modified_gmt"] = pd.to_datetime(pages_df["modified_gmt"])
    pages_df = pages_df.drop(["date", "modified"], axis=1)

    pages_df["link_locale"] = pages_df["link"].apply(extract_locale)

    pages_df["title.text"] = pages_df["title.rendered"].apply(extract_html_text)
    pages_df["excerpt.text"] = pages_df["excerpt.rendered"].apply(extract_html_text)

    tqdm.pandas(desc="Parsing post content")
    pages_df["content.bs"] = pages_df["content.rendered"].progress_apply(parse_html)

    tqdm.pandas(desc="Extracting post content")
    pages_df[
        ["content.text", "links.internal", "links.external", "embeds", "images"]
    ] = pages_df.progress_apply(
        lambda r: extract_content_data(r["content.bs"], r["link"]), axis=1
    )

    pages_df = pages_df[pages_df.columns.intersection(EXPORT_COLUMNS)]
    pages_df = pages_df.rename(columns=RENAME_COLUMNS, errors="ignore")

    link_registry.add_linkables(
        "pages", pages_df["link"].to_list(), pages_df.index.to_list()
    )

    return pages_df



import logging
from pathlib import Path
from typing import Optional

from bs4 import BeautifulSoup


def load_scrape(
    scrape_urls_files: dict[str, Path], link: str
) -> Optional[BeautifulSoup]:
    """Find, load and parse the scrape of a link.

    Args:
        scrape_urls_files: A dictionary of site URLs to scrape file paths
        link: The link of the page to find.

    Returns:
        The parsed document or None if it could not be found.
    """
    path = scrape_urls_files.get(link)

    if path is None:
        logging.debug(f'Unable to find scrape file for "{link}"')
        return None

    html_raw = path.read_text()
    doc = BeautifulSoup(html_raw, "lxml")
    return doc



import json
from urllib.parse import urlsplit, urlunsplit


def get_by_id(value, id):
    """Utility function to retrieve a value by and ID in a list of dicts.

    Args:
        value: the dict to process
        id: the id to get

    Returns:
        The matching value or None if no match is found
    """
    if value is None:
        return None
    for val in value:
        if "id" in val.keys() and val["id"] == id:
            return val
    return None


# Neat code part from https://codereview.stackexchange.com/questions/13027/joini
# ng-url-path-components-intelligently
def url_path_join(*parts):
    """Normalize url parts and join them with a slash."""
    schemes, netlocs, paths, queries, fragments = zip(
        *(urlsplit(part) for part in parts)
    )
    scheme = first(schemes)
    netloc = first(netlocs)
    path = "/".join(x.strip("/") for x in paths if x)
    query = first(queries)
    fragment = first(fragments)
    return urlunsplit((scheme, netloc, path, query, fragment))


def first(sequence, default=""):
    """Return the first element of an iterable sequence or a default value.

    Args:
        sequence: an iterable sequence.
        default: the value to return if the sequence is empty.

    Returns:
        The first element of an iterable sequence or a default value.
    """
    return next((x for x in sequence if x), default)


def get_content_as_json(response_obj):
    """Returns a json value even if a BOM is present in UTF-8 text.

    When a BOM is present (see issue #2), UTF-8 is not properly decoded by
    Response.json() method.


    Args:
        response_obj: a requests Response instance

    Returns:
        a decoded json object (list or dict)
    """
    if response_obj.content[:3] == b"\xef\xbb\xbf":  # UTF-8 BOM
        content = response_obj.content.decode("utf-8-sig")
        return json.loads(content)
    else:
        try:
            return response_obj.json()
        except:  # noqa: E722
            return {}



from abc import ABC, abstractmethod

from bs4 import BeautifulSoup, PageElement, Tag
from langcodes import Language

from wpextract.parse.translations._resolver import TranslationLink
from wpextract.util.str import squash_whitespace


class LangPicker(ABC):
    """Abstract class of a language picker style.

    Support for a new language picker can be added by creating a new class inheriting from this one.
    """

    page_doc: BeautifulSoup
    """The document to extract the language picker from."""
    root_el: Tag
    """The root element of the language picker, populated if [`LangPicker.matches`][wpextract.parse.translations.LangPicker.matches] is succesful."""
    translations: list[TranslationLink]
    """A list of translation links, populated by calling [`LangPicker.add_translation`][wpextract.parse.translations.LangPicker.add_translation] within [`LangPicker.extract`][wpextract.parse.translations.LangPicker.extract]."""
    current_language: Language
    """The current language of the page, populated by calling [`LangPicker.set_current_lang`][wpextract.parse.translations.LangPicker.set_current_lang] within [`LangPicker.extract`][wpextract.parse.translations.LangPicker.extract]."""

    def __init__(self, page_doc: BeautifulSoup):
        """Inits a language picker searcher.

        Args:
            page_doc: The document to extract a language picker from.
        """
        self.page_doc = page_doc
        self.translations = []

    def matches(self) -> bool:
        """Checks if this picker can extract from the document.

        Returns:
            If the page uses this type of matcher.

        Raises:
            TypeError: If the root element that has been retrieved is not a tag,
                or has 0 children.
                This may happen if it accidentally retrieves a text node.
        """
        root = self.get_root()

        if root is None:
            return False

        if type(root) is Tag or len(root.children) == 1:
            self.root_el = root
            return True
        else:
            raise TypeError(f"Root is not a tag, is {type(root)}")

    @abstractmethod
    def get_root(self) -> PageElement:
        """Retrieve the root element of the translation picker.

        Using the [`LangPicker.page_doc`][wpextract.parse.translations.LangPicker.page_doc] attribute (a [`bs4.BeautifulSoup`][bs4.BeautifulSoup] object representing the whole page), the root element of the picker shoudl be found and returned.

        Returns:
            The root element, or None if this picker is not found on the page.
        """
        pass

    @abstractmethod
    def extract(self) -> None:
        """Extract the current language and translations from the doc."""
        pass

    def set_current_lang(self, lang: str) -> None:
        """Set the language of this doc.

        Args:
            lang: The locale string
        """
        self.current_language = Language.get(lang, normalize=True)

    def add_translation(self, href: str, lang: str) -> None:
        """Add a translation from the picker.

        Args:
            href: The link to the translated page.
            lang: The provided language code.
        """
        self.translations.append(
            TranslationLink(text=None, href=href, destination=None, lang=lang)
        )


class PolylangWidget(LangPicker):
    """Language picker from the plugin Polylang used as a widget.

    [WordPress plugin page](https://wordpress.org/plugins/polylang/)

    Has the structure:

    ```html
    <div id="polylang" class="widget widget_polylang">
        <ul>
            <li class="lang-item lang-item-en current-lang lang-item-first">
                <a lang="en-US" hreflang="en-US" href="URL">\
                <img><span>English</span></a>
            </li>
            <li class="lang-item lang-item-20 lang-item-fr">
                <a lang="fr-FR" hreflang="fr-FR" href="URL">\
                <img><span>Français</span></a>
            </li>
            ...
        </ul>
    </div>
    ```
    """

    def get_root(self) -> PageElement:
        """Get the root element.

        Returns:
            The .widget_polylang tag
        """
        return self.page_doc.select_one(".widget_polylang")

    def extract(self) -> None:
        """Extract the translation links.

        Retrieves the current language by checking the ``lang`` attribute of a
        link within an `li` with the class `.lang-item.current-lang`.

        Retrieves translations by searching for links within ``.lang-item`` unless they:
        * have the ``.no-translation`` class - this language is missing and the link \
            will usually go to the language's homepage
        * have the ``.current-lang`` class - this is the current language.
        """
        current_lang = self.root_el.select_one(".lang-item.current-lang a")
        self.set_current_lang(current_lang["lang"])

        other_langs = self.root_el.select(
            ".lang-item:not(.no-translation):not(.current-lang) a"
        )

        for lang_a in other_langs:
            href = lang_a["href"]
            lang = lang_a["lang"]

            self.add_translation(href, lang)


class PolylangCustomDropdown(LangPicker):
    """Language picker for an in-the-wild version of polylang.

    This was implemented to support a specific site of interest.

    Finds language pickers in the form:

    ```html
    <div class="header-lang_switcher switcher-ltr">
        <div class="current-lang-switcher">
            <img><span>en</span>
        </div>
        <ul>
            <li class="lang-item lang-item-fr lang-item-first">
                <a lang="fr-FR" hreflang="fr-FR" href="">Français</a>
            </li>
        </ul>
    </div>
    ```
    """

    def get_root(self) -> PageElement:
        """Get the root element of the picker.

        Returns:
            The lang switcher root.
        """
        return self.page_doc.select_one(".header-lang_switcher")

    def extract(self) -> None:
        """Extract the translation links.

        Retrieves all links in a ``.lang-item`` unless they:
        * have the ``.no-translation`` class
        * have the ``current-lang`` class
        """
        current_lang = self.root_el.select_one(".current-lang-switcher")
        self.set_current_lang(squash_whitespace(current_lang.get_text()))

        other_langs = self.root_el.select(".lang-item:not(.no-translation) a")

        for lang_a in other_langs:
            href = lang_a["href"]
            lang = lang_a["lang"]

            self.add_translation(href, lang)



import logging
from pathlib import Path
from typing import Optional

from pandas import DataFrame

from wpextract.extractors.categories import load_categories
from wpextract.extractors.data.links import LinkRegistry
from wpextract.extractors.io import export_df
from wpextract.extractors.media import load_media
from wpextract.extractors.pages import load_pages
from wpextract.extractors.posts import (
    ensure_translations_undirected,
    load_posts,
    resolve_post_links,
    resolve_post_translations,
)
from wpextract.extractors.tags import load_tags
from wpextract.extractors.users import load_users
from wpextract.parse.translations import PickerListType
from wpextract.scrape.crawler import ScrapeCrawl
from wpextract.util.file import prefix_filename


class WPExtractor:
    """Manages the extraction of data from a WordPress site."""

    json_root: Path
    scrape_root: Optional[Path]
    json_prefix: Optional[str]
    translation_pickers: Optional[PickerListType]

    link_registry: LinkRegistry
    """Registry of known URLs and their corresponding data items."""

    posts: Optional[DataFrame]
    """DataFrame of extracted posts."""
    media: Optional[DataFrame]
    """DataFrame of extracted media."""
    tags: Optional[DataFrame]
    """DataFrame of extracted tags."""
    categories: Optional[DataFrame]
    """DataFrame of extracted categories."""
    users: Optional[DataFrame]
    """DataFrame of extracted users."""
    pages: Optional[DataFrame]
    """DataFrame of extracted pages."""

    scrape_url_mapping: dict[str, Path]

    def __init__(
        self,
        json_root: Path,
        scrape_root: Optional[Path] = None,
        json_prefix: Optional[str] = None,
        translation_pickers: Optional[PickerListType] = None,
    ):
        """Create a new extractor.

        Args:
            json_root: Path to directory of JSON files
            scrape_root: Path to scrape directory
            json_prefix: Prefix of files in ``json_root``
            translation_pickers: Supply a custom list of translation pickers
        """
        self.json_root = json_root
        self.scrape_root = scrape_root
        self.json_prefix = json_prefix
        self.link_registry = LinkRegistry()
        self.translation_pickers = translation_pickers

    def extract(self) -> None:
        """Perform the extraction."""
        logging.info("Beginning extraction")
        logging.info("Beginning scrape crawl")
        self._crawl_scrape()
        logging.info("Beginning media extraction")
        self._extract_media()
        logging.info("Beginning post extraction")
        self._extract_posts()
        logging.info("Beginning page extraction")
        self._extract_pages()
        logging.info("Beginning tag extraction")
        self._extract_tags()
        logging.info("Beginning categories extraction")
        self._extract_categories()
        logging.info("Beginning user extraction")
        self._extract_users()
        logging.info("Beginning post link matching")
        self._resolve_post_links()
        logging.info("Extraction complete")

    def _prefix_filename(self, file_name):
        return prefix_filename(file_name, self.json_prefix)

    def _crawl_scrape(self):
        if self.scrape_root is None:
            logging.info("No scrape root specified, skipping")
            self.scrape_url_mapping = {}
            return

        crawl = ScrapeCrawl(self.scrape_root)
        crawl.crawl()
        self.scrape_url_mapping = crawl.get_link_abs_path()

    def _extract_posts(self):
        json_file = self.json_root / self._prefix_filename("posts.json")
        self.posts = load_posts(
            path=json_file,
            link_registry=self.link_registry,
            scrape_urls_files=self.scrape_url_mapping,
            translation_pickers=self.translation_pickers,
        )

    def _extract_media(self):
        json_file = self.json_root / self._prefix_filename("media.json")
        self.media = load_media(json_file, self.link_registry)

    def _extract_tags(self):
        json_file = self.json_root / self._prefix_filename("tags.json")
        self.tags = load_tags(json_file, self.link_registry)

    def _extract_categories(self):
        json_file = self.json_root / self._prefix_filename("categories.json")
        self.categories = load_categories(json_file, self.link_registry)

    def _extract_users(self):
        json_file = self.json_root / self._prefix_filename("users.json")
        self.users = load_users(json_file)

    def _extract_pages(self):
        json_file = self.json_root / self._prefix_filename("pages.json")
        self.pages = load_pages(json_file, self.link_registry)

    def _resolve_post_links(self):
        self.posts = resolve_post_links(self.link_registry, self.posts)
        if "translations" in self.posts.columns:
            self.posts = resolve_post_translations(self.link_registry, self.posts)
            self.posts = ensure_translations_undirected(self.posts)
            self.posts = resolve_post_translations(self.link_registry, self.posts)

    def export(self, out_dir: Path) -> None:
        """Save scrape results!

        Args:
            out_dir: Path to output directory
        """
        logging.info("Beginning export")
        export_df(self.posts, out_dir / self._prefix_filename("posts.json"))
        export_df(self.pages, out_dir / self._prefix_filename("pages.json"))
        export_df(self.media, out_dir / self._prefix_filename("media.json"))
        export_df(self.tags, out_dir / self._prefix_filename("tags.json"))
        export_df(self.categories, out_dir / self._prefix_filename("categories.json"))
        export_df(self.users, out_dir / self._prefix_filename("users.json"))
        logging.info("Completed export")



import logging
from typing import Optional

import pandas as pd
from bs4 import BeautifulSoup

import wpextract.parse.translations._pickers as pickers

DEFAULT_PICKERS = [pickers.PolylangWidget, pickers.PolylangCustomDropdown]
PickerListType = list[type[pickers.LangPicker]]

PageTranslationData = pd.Series


def extract_translations(
    page_doc: Optional[BeautifulSoup],
    link: str,
    translation_pickers: Optional[PickerListType],
) -> PageTranslationData:
    """Get a list of URLs linked as translations.

    Args:
        page_doc: The full scrape document
        link: The link to the post
        translation_pickers: A list of translation pickers to use

    Returns:
        The doc's language and list of translation links
    """
    if translation_pickers is None:
        translation_pickers = DEFAULT_PICKERS

    if page_doc is None:
        return pd.Series([None, []])
    for picker_class in translation_pickers:
        picker = picker_class(page_doc)

        if not picker.matches():
            continue

        picker.extract()

        return pd.Series([picker.current_language, picker.translations])

    logging.debug(
        f'No translation pickers matched "{link}", unable to extract translations.'
    )

    return pd.Series([None, []])



import urllib.parse
from typing import Optional

from langcodes import Language, tag_is_valid

from wpextract.util.str import remove_ends

EXCLUDED_TAGS = {"tag"}  # Tagoi


def extract_locale(link: str) -> Optional[str]:
    """Extract a locale string from a URL.

    Returns the string standardised into a common format.

    Examples:
        >>> extract_locale("https://example.org/news/foo")
        None

        >>> extract_locale("https://example.org/en/foo")
        "en"

        >>> extract_locale("https://example.org/en-gb/foo")
        "en-GB"

    Args:
        link: A URL

    Returns:
        A standardised locale.
    """
    parsed_link = urllib.parse.urlparse(link)
    path_stripped = remove_ends(parsed_link.path, "/")
    path_parts = path_stripped.split("/")
    if tag_is_valid(path_parts[0]) and path_parts[0] not in EXCLUDED_TAGS:
        lang = Language.get(path_parts[0], normalize=True)
        if lang.extensions is not None:
            # Language tag extensions are very unlikely to be used
            # Probably a mis-parse
            return None

        lang.prefer_macrolanguage()
        return lang.simplify_script().to_tag()

    return None



import re

from bs4 import BeautifulSoup

from wpextract.util.str import squash_whitespace

PROBABLY_HTML = re.compile(r"<|&\S+;")


def parse_html(html: str) -> BeautifulSoup:
    """Helper to parse HTML into a BeautifulSoup document.

    Args:
        html: an HTML document as a string.

    Returns:
        The parsed document
    """
    return BeautifulSoup(html, "lxml")


def extract_html_text(html: str) -> str:
    """Extract text from an HTML document.

    Will skip parsing if the document does not contain "<" or an HTML-entity-like
    sequence ("&" followed by some characters and a ";")

    Args:
        html: A string containing HTML.

    Returns:
        The extracted text from the document
    """
    if PROBABLY_HTML.search(html) is None:
        return html

    return squash_whitespace(parse_html(html).get_text())



import urllib.parse
from typing import Optional

from bs4 import BeautifulSoup, SoupStrainer

self_url_strainer = SoupStrainer(["head", "link", "meta"])


def _is_url_valid(url: str) -> bool:
    parsed = urllib.parse.urlparse(url)

    return all([parsed.scheme, parsed.netloc])


def get_link_canonical(doc: BeautifulSoup) -> Optional[str]:
    """Get the canonical link if present.

    Args:
        doc: The document to search

    Returns:
        The canonical link or None if not present.
    """
    link_canonical = doc.head.find("link", rel="canonical")

    if link_canonical is None or not link_canonical.has_attr("href"):
        return None

    url = link_canonical["href"]

    return url if _is_url_valid(url) else None


def get_og_url(doc: BeautifulSoup) -> Optional[str]:
    """Get the meta ``og:url`` if present.

    Args:
        doc: The doc to search

    Returns:
        The og:url or None if not present.
    """
    og_url = doc.head.find("meta", property="og:url")

    if og_url is None or not og_url.has_attr("content"):
        return None

    url = og_url["content"]

    return url if _is_url_valid(url) else None


def extract_self_url(doc: BeautifulSoup) -> Optional[str]:
    """Extract the document's URL from meta tags.

    Args:
        doc: The doc to search

    Returns:
        The document's URL or None if it could not be found.
    """
    link_canonical = get_link_canonical(doc)
    if link_canonical is not None:
        return link_canonical

    og_url = get_og_url(doc)
    if og_url is not None:
        return og_url

    return None



import logging
from pathlib import Path
from typing import Any, Callable, Optional

from wpextract.download.exceptions import WordPressApiNotV2
from wpextract.download.exporter import Exporter
from wpextract.download.requestsession import RequestSession
from wpextract.download.wpapi import WPApi


class WPDownloader:
    """Manages the download of data from a WordPress site."""

    def __init__(
        self,
        target: str,
        out_path: Path,
        data_types: list[str],
        session: Optional[RequestSession] = None,
        json_prefix: Optional[str] = None,
    ):
        """Initializes the WPDownloader object.

        Args:
            target: the target WordPress site URL
            out_path: the output path for the downloaded data
            data_types: set of data types to download
            session : request session. Will be created from default constructor if not provided.
            json_prefix: prefix to prepend to JSON file names
        """
        self.target = target
        self.out_path = out_path
        self.data_types = data_types
        self.session = session if session else RequestSession()
        self._test_session()
        self.scanner = WPApi(self.target, session=self.session)
        self.json_prefix = json_prefix

    def _test_session(self):
        try:
            self.session.get(self.target)
            logging.info("Connected successfully")
        except Exception as e:
            logging.error("Failed to connect to the server")
            raise e

    def download(self):
        """Download and export the requested data lists."""
        if "users" in self.data_types:
            self._list_obj(WPApi.USER)
        if "tags" in self.data_types:
            self._list_obj(WPApi.TAG)
        if "categories" in self.data_types:
            self._list_obj(WPApi.CATEGORY)
        if "posts" in self.data_types:
            self._list_obj(WPApi.POST)
        if "pages" in self.data_types:
            self._list_obj(WPApi.PAGE)
        if "comments" in self.data_types:
            self._list_obj(WPApi.COMMENT)
        if "media" in self.data_types:
            self._list_obj(WPApi.MEDIA)

    def download_media_files(self, session: RequestSession, dest: Path):
        """Download site media files.

        Args:
            session: the request session to use
            dest: destination directory for media
        """
        logging.info("Pulling media URLs")
        media, slugs = self.scanner.get_media_urls("all", cache=True)

        if len(media) == 0:
            logging.warning("No media found corresponding to the criteria")
            return
        logging.info(f"{len(media)} media URLs found")

        number_dl = Exporter.download_media(session, media, dest)
        logging.info(f"Downloaded {number_dl} media files")

    def _get_fetch_or_list_type(self, obj_type, plural=False):
        """Returns a dict containing all necessary metadata about the obj_type to list and fetch data.

        Args:
            obj_type: the type of the object
            plural: whether the name must be plural or not
        """
        export_func = None
        obj_name = ""
        if obj_type == WPApi.USER:
            export_func = Exporter.export_users
            obj_name = "Users" if plural else "User"
        elif obj_type == WPApi.TAG:
            export_func = Exporter.export_tags
            obj_name = "Tags" if plural else "Tag"
        elif obj_type == WPApi.CATEGORY:
            export_func = Exporter.export_categories
            obj_name = "Categories" if plural else "Category"
        elif obj_type == WPApi.POST:
            export_func = Exporter.export_posts
            obj_name = "Posts" if plural else "Post"
        elif obj_type == WPApi.PAGE:
            export_func = Exporter.export_pages
            obj_name = "Pages" if plural else "Page"
        elif obj_type == WPApi.COMMENT:
            export_func = Exporter.export_comments_interactive
            obj_name = "Comments" if plural else "Comment"
        elif obj_type == WPApi.MEDIA:
            export_func = Exporter.export_media
            obj_name = "Media"

        return {
            "export_func": export_func,
            "obj_name": obj_name,
        }

    def _list_obj(self, obj_type, start=None, limit=None, cache=True):
        prop = self._get_fetch_or_list_type(obj_type, plural=True)
        logging.info(f"Downloading {prop['obj_name']}")

        try:
            kwargs = {}
            if obj_type == WPApi.POST:
                kwargs = {"comments": False}
            obj_list = self.scanner.get_obj_list(
                obj_type, start, limit, cache, kwargs=kwargs
            )

            WPDownloader.export_decorator(
                export_func=prop["export_func"],
                file_name=prop["obj_name"].lower(),
                json_path=self.out_path,
                json_prefix=self.json_prefix,
                values=obj_list,
            )
        except WordPressApiNotV2:
            logging.error("The API does not support WP V2")
        except OSError as e:
            logging.error(f"Could not open {e.filename} for writing")
        logging.info(f"Completed downloading {prop['obj_name']}")

    @staticmethod
    def export_decorator(
        export_func: Callable,
        file_name: str,
        json_path: Path,
        json_prefix: str,
        values: Any,
        kwargs: Optional[dict] = None,
    ) -> None:
        """Call the export function with a constructed filename.

        Args:
            export_func: the function to run
            file_name: the name of the export file
            json_path: the path to the output directory
            json_prefix: a prefix for the export file
            values: data to be exported
            kwargs: arguments to pass to the export function
        """
        kwargs = kwargs or {}

        filename = file_name + ".json"
        if json_prefix is not None:
            filename = json_prefix + "-" + filename

        json_file = json_path / filename
        export_func(values, json_file, **kwargs)



from pathlib import Path
from typing import Optional

import click
from click import Choice
from click_option_group import optgroup

from wpextract import WPDownloader
from wpextract.cli._shared import (
    CMD_ARGS,
    empty_directory,
    logging_options,
    setup_logging,
    setup_tqdm_redirect,
)
from wpextract.download import RequestSession
from wpextract.util.str import ensure_prefixes, ensure_suffix

dl_types = ["categories", "media", "pages", "posts", "tags", "users"]


def data_type_opts(cmd_func):
    opts = [
        click.option(
            f"--{dl_type}/--no-{dl_type}",
            default=True,
            help=f"Enable or disable downloading {dl_type}",
        )
        for dl_type in dl_types
    ]

    for opt in opts:
        cmd_func = opt(cmd_func)
    return cmd_func


def validate_wait(ctx, param, value):
    if ctx.params.get("wait") is None and value is not False:
        raise click.BadParameter("cannot be used unless --wait/-w is also set.")
    return value


@click.command(short_help="Download a WordPress site.", **CMD_ARGS)
@click.argument("target", type=str)
@click.argument("out_json", type=click.Path(), callback=empty_directory)
@click.option(
    "--media-dest",
    type=click.Path(),
    callback=empty_directory,
    required=False,
    help="Path to a directory to download media files to, skipped if not supplied",
    metavar="DIRECTORY",
)
@click.option(
    "-P", "--json-prefix", type=str, help="Prefix to add to output file names"
)
@click.option(
    "--skip-type",
    "skip_types",
    type=Choice(dl_types, case_sensitive=False),
    default=[],
    multiple=True,
    help="Don't download the provided types. All others will be downloaded, default is to download all.",
)
@optgroup.group("authentication")
@optgroup.option("--proxy", type=str, help="Proxy server for requests")
@optgroup.option(
    "--auth",
    type=str,
    help='HTTP Basic credentials for requests (format "username:password")',
)
@optgroup.option(
    "--cookies",
    type=str,
    help='Cookies for requests (format "cookie1=foo; cookie2=bar")',
)
@optgroup.group("request behaviour")
@optgroup.option(
    "--timeout",
    type=int,
    default=30,
    help="Timeout for request in seconds",
    show_default=True,
)
@optgroup.option(
    "-w",
    "--wait",
    type=int,
    help="Time to wait between requests in seconds. Does not affect retries.",
    is_eager=True,  # to permit --random-wait validation
)
@optgroup.option(
    "--random-wait",
    is_flag=True,
    help="Randomly varies the time between requests to between 0.5 and 1.5 times the number of seconds set by --wait",
    callback=validate_wait,
)
@optgroup.option(
    "--max-retries",
    type=int,
    default=10,
    help="Maximum number of retries before giving up",
    show_default=True,
)
@optgroup.option(
    "--backoff-factor",
    type=float,
    default=0.1,
    help="Factor to apply delaying retries. Default will sleep for 0.0, 0.2, 0.4, 0.8,...",
    show_default=True,
)
@optgroup.option(
    "--max-redirects",
    type=int,
    default=20,
    help="Maximum number of redirects before giving up",
    show_default=True,
)
@logging_options
def download(
    target: str,
    out_json: Path,
    media_dest: Optional[Path],
    json_prefix: Optional[str],
    skip_types: list[str],
    proxy: Optional[str],
    auth: Optional[str],
    cookies: Optional[str],
    timeout: int,
    wait: Optional[int],
    random_wait: bool,
    max_retries: int,
    backoff_factor: float,
    max_redirects: int,
    log: Optional[Path],
    verbose: bool,
):
    """Download a site's content using the WordPress REST API.

    TARGET is the base path of the WordPress installation, e.g. "https://example.org/"

    OUT_JSON is the directory to output the downloaded JSON to. It must be an existing empty directory or a non-existent directory which will be created.
    """
    setup_logging(verbose, log)

    types_to_dl = set(dl_types) - set(skip_types)

    target = ensure_prefixes(target, ("http://", "https://"), "http://")
    target = ensure_suffix(target, "/")

    if auth is not None:
        auth_list = auth.split(":")
        if len(auth_list) == 1:
            auth = (auth_list[0], "")
        elif len(auth_list) >= 2:
            auth = (auth_list[0], ":".join(auth_list[1:]))

    session = RequestSession(
        proxy=proxy,
        cookies=cookies,
        authorization=auth,
        timeout=timeout,
        wait=wait,
        random_wait=random_wait,
        max_retries=max_retries,
        backoff_factor=backoff_factor,
        max_redirects=max_redirects,
    )

    with setup_tqdm_redirect(log is None):
        downloader = WPDownloader(
            target=target,
            out_path=out_json,
            data_types=list(types_to_dl),
            session=session,
            json_prefix=json_prefix,
        )

        downloader.download()

        if media_dest is not None:
            downloader.download_media_files(session, media_dest)



import logging
import re
from dataclasses import dataclass
from typing import Optional

from wpextract.extractors.data.links import Linkable, LinkRegistry


@dataclass
class MediaUse:
    """An instance of media being used."""

    src: str
    alt: str
    caption: Optional[str]


@dataclass
class ResolvableMediaUse(MediaUse):
    """An instance of media that can be resolved against known media."""

    destination: Optional[Linkable] = None


def resolve_image(
    registry: LinkRegistry, image: ResolvableMediaUse
) -> ResolvableMediaUse:
    """Resolve the internal links of a media use.

    Args:
        registry: A filled link registry
        image: A media use

    Returns:
        The media use with link data resolved.
    """
    if image.destination is not None:
        return image

    if "wp-content" not in image.src:
        return image

    # Remove dimensions from image URL
    # e.g. test-image-300x200.jpg -> test-image.jpg
    src = re.sub(r"-\d{3,4}x\d{3,4}\.", ".", image.src)

    linkable = registry.query_link(src)

    if linkable is None:
        logging.debug(f'Could not resolve image "{src}"')
        return image

    image.destination = linkable

    return image


def resolve_images(
    registry: LinkRegistry, images: list[ResolvableMediaUse]
) -> list[ResolvableMediaUse]:
    """Resolve the internal links of a list of media uses.

    Args:
        registry: A filled link registry
        images: A list of media uses

    Returns:
        The list of media uses with link data resolved.
    """
    return [resolve_image(registry, image) for image in images]



import re

ADJACENT_NEWLINE_REGEX = re.compile("\n{2,}")
WHITESPACE_LINE_STARTING_REGEX = re.compile(r"^\s+", re.MULTILINE)


def remove_prefix(text: str, prefix: str):
    """If the text starts with the prefix, remove it.

    If the prefix is not present, the original string will be returned.

    Args:
        text: The string to process
        prefix: The prefix to remove

    Returns:
        The text with the given prefix removed.
    """
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def remove_suffix(text: str, suffix: str):
    """If the text ends with the suffix, remove it.

    If the suffix is not present, the original string will be returned.

    Args:
        text: The string to process
        suffix: The suffix to remove

    Returns:
        The text with the given prefix removed.
    """
    if text.endswith(suffix):
        return text[: -len(suffix)]
    return text


def remove_ends(text: str, affix: str):
    """If the text starts or ends with the affix, remove it.

    If the affix is not present, the original string will be returned.

    Args:
        text: The string to process
        affix: The prefix and suffix to remove

    Returns:
        The text with the prefix and suffix removed.
    """
    return remove_prefix(remove_suffix(text, affix), affix)


def ensure_prefixes(text: str, prefixes: tuple[str], default: str):
    """Ensure the text starts with one of several prefixes, or prepend a default.

    Args:
        text: The string to process
        prefixes: acceptable prefixes to the string
        default: a prefix to prepend if none of the prefixes are present

    Returns:
        The prefixed string, or original string if no prefixes are present.
    """
    if not text.startswith(prefixes):
        return default + text
    return text


def ensure_prefix(text: str, prefix: str) -> str:
    """Ensure the text starts with a prefix.

    If it already starts with the prefix, the original string will be returned.

    Args:
        text: The string to process
        prefix: The prefix to ensure

    Returns:
        The text with the given prefix ensured.
    """
    if not text.startswith(prefix):
        return prefix + text
    return text


def ensure_suffix(text: str, suffix: str) -> str:
    """Ensure the text ends with the suffix.

    If it already ends with the suffix, the original string will be returned.

    Args:
        text: The string to process
        suffix: The suffix to add if not already present.

    Returns:
        The text with the given suffix.
    """
    if not text.endswith(suffix):
        return text + suffix
    return text


def squash_whitespace(string: str) -> str:
    """Reduces unnecessary whitespace in (multiline) strings.

    Replaces multiple adjacent newlines with a single newline. Removes whitespace at the
    start of a line. Removes whitespace at the start and end of the string.

    Args:
        string: A string to strip.

    Returns:
        The stripped strings.
    """
    return WHITESPACE_LINE_STARTING_REGEX.sub(
        "", ADJACENT_NEWLINE_REGEX.sub("\n", string)
    ).strip()



from typing import Optional


def prefix_filename(file_name: str, prefix: Optional[str]):
    """Adds an optional prefix to the file name.

    If the prefix is ``None``, no prefix will be added.

    Args:
        file_name: The base file name.
        prefix: A prefix, or ``None``.

    Returns:
        The file name with prefix.
    """
    if prefix is None:
        prefix = ""

    return prefix + file_name



import copy
import logging
from urllib.parse import urljoin, urlparse, urlunparse

import pandas as pd
from bs4 import BeautifulSoup, Comment, NavigableString

from wpextract.extractors.data.images import MediaUse, ResolvableMediaUse
from wpextract.extractors.data.links import Link, ResolvableLink
from wpextract.extractors.media import get_caption
from wpextract.util.str import squash_whitespace

EXCLUDED_CONTENT_TAGS = {"figcaption"}
NEWLINE_TAGS = {"br", "p"}


InternalLinks = list[ResolvableLink]
ExternalLinks = list[Link]


def extract_links(
    doc: BeautifulSoup, self_link: str
) -> tuple[InternalLinks, ExternalLinks]:
    """Get the internal and external links of the document.

    Args:
        doc: A parsed document
        self_link: The link of the page

    Returns:
        A list of internal and a list of external link objects.
    """
    internal_links = []
    external_links = []

    links = doc.find_all("a")

    self_link_parsed = urlparse(self_link)

    for link in links:
        if not link.has_attr("href"):
            external_links.append(Link(squash_whitespace(link.get_text()), None))
            continue

        href_parsed = urlparse(urljoin(self_link, link["href"]))
        if href_parsed.netloc == self_link_parsed.netloc:
            internal_links.append(
                ResolvableLink(
                    text=squash_whitespace(link.get_text()),
                    href=urlunparse(href_parsed),
                    destination=None,
                )
            )
        else:
            external_links.append(
                Link(text=squash_whitespace(link.get_text()), href=link["href"])
            )

    return internal_links, external_links


Embeds = list[str]


def extract_embeds(doc: BeautifulSoup) -> Embeds:
    """Get a list of URLs embedded in the page.

    Args:
        doc: A parsed document.

    Returns:
        A list of URLs
    """
    return [iframe["src"] for iframe in doc.find_all("iframe")]


Images = list[MediaUse]


def extract_images(doc: BeautifulSoup, self_link: str) -> Images:
    """Get a list of images in the document.

    Args:
        doc: A parsed document
        self_link: The link of the page.

    Returns:
        A list of MediaUse objects. Internal images are returned as ResolvableMedia.
    """
    images = doc.find_all("img")

    media_uses = []

    self_link_parsed = urlparse(self_link)

    for img in images:
        if not img.has_attr("src"):
            logging.warning(f"Image without source in {self_link}")
            media_uses.append(
                MediaUse(src="", alt=img.get("alt"), caption=get_caption(img))
            )
            continue

        src_parsed = urlparse(urljoin(self_link, img["src"]))

        media_data = {
            "src": urlunparse(src_parsed),
            "alt": img.get("alt"),
            "caption": get_caption(img),
        }

        if src_parsed.netloc == self_link_parsed.netloc:
            media_uses.append(ResolvableMediaUse(**media_data))
        else:
            media_uses.append(MediaUse(**media_data))

    return media_uses


def _get_text(doc: BeautifulSoup) -> str:
    """Custom function to get document text.

    Extracts text from all elements, inserting newlines for <p> and <br> tags.
    """
    text = ""
    for e in doc.descendants:
        # Comments are a subtype of NavigableString, they need to be excluded
        if isinstance(e, NavigableString) and not isinstance(e, Comment):
            text += e
        elif e.name in NEWLINE_TAGS:
            text += "\n"
    return text


def extract_content_data(doc: BeautifulSoup, self_link: str) -> pd.Series:
    """Extract the links, embeds, images and text content of the document.

    Args:
        doc: A parsed document.
        self_link: The URL of the page.

    Returns:
        A series with the text, internal links, external links, embeds and images.
    """
    internal_links, external_links = extract_links(doc, self_link)
    embeds = extract_embeds(doc)
    images = extract_images(doc, self_link)

    doc_c = copy.copy(doc)
    for child in doc_c.descendants:
        if type(child) == NavigableString:
            continue

        if child.name in EXCLUDED_CONTENT_TAGS:
            child.extract()

    content_text = squash_whitespace(_get_text(doc_c))

    return pd.Series([content_text, internal_links, external_links, embeds, images])



from dataclasses import dataclass

from langcodes import Language

from wpextract.extractors.data.links import ResolvableLink


@dataclass
class TranslationLink(ResolvableLink):
    """A link to an alternative version of this article in a different language."""

    lang: str
    """Raw language code."""

    @property
    def language(self) -> Language:
        """Parsed and normalized language. Populated automatically post-init."""
        return Language.get(self.lang, normalize=True)



from ._extractor import PickerListType as PickerListType
from ._extractor import extract_translations as extract_translations
from ._pickers import LangPicker as LangPicker
from ._resolver import TranslationLink as TranslationLink



import logging
from typing import Optional
from urllib.parse import urlparse, urlunparse

from wpextract.extractors.data.links import LinkRegistry, ResolvableLink
from wpextract.util.str import remove_ends


def resolve_link(
    registry: LinkRegistry, link: ResolvableLink
) -> Optional[ResolvableLink]:
    """Find the linkable item from the registry the link refers to.

    Will not resolve a link if it already has a destination.

    Args:
        registry: A link registry.
        link: A resolvable link

    Returns:
        The resolved link.
    """
    if link.destination is not None:
        return link

    href_parsed = urlparse(link.href)
    if "preview_id" in href_parsed.query:
        href_parsed = href_parsed._replace(query="")
        href = urlunparse(href_parsed)
    else:
        href = link.href

    linkable = registry.query_link(href)

    if linkable is None:
        # Heuristic to fix case when category slug has been removed from URL
        path_parts = remove_ends(href_parsed.path, "/").split("/")

        lang = None
        # Case /fr/category/article/
        if len(path_parts) == 3 and (len(path_parts[0]) == 2):
            lang = path_parts.pop(0)

        # Case /category/an-article-slug/
        if len(path_parts) == 2:
            path_parts.pop(0)
            if lang is not None:
                path_parts.insert(0, lang)
            whole_path = f"/{'/'.join(path_parts)}/"
            href_parsed = href_parsed._replace(path=whole_path)
            url = urlunparse(href_parsed)
            linkable = registry.query_link(url)

            if linkable is None:
                logging.debug(
                    f'Could not resolve with category removal heuristic: "{url}"'
                )

    if linkable is None:
        logging.debug(f'Could not resolve link "{href}"')
        return link

    link.destination = linkable

    return link


def resolve_links(
    registry: LinkRegistry, links: list[ResolvableLink]
) -> list[ResolvableLink]:
    """Resolve a list of links against the link registry.

    Args:
        registry: A link registry
        links: A list of resolvable links

    Returns:
        The list of resolvable links with destinations.
    """
    return [resolve_link(registry, link) for link in links]



from dataclasses import dataclass
from typing import Optional


@dataclass
class Linkable:
    """An item which can be linked to."""

    link: str
    data_type: str
    idx: str


@dataclass
class Link:
    """A link to a URL."""

    text: Optional[str]
    href: Optional[str]


@dataclass
class ResolvableLink(Link):
    """A link to a URL which can be looked up against known links."""

    destination: Optional[Linkable]


class LinkRegistry:
    """A collection of all known links on the site."""

    links: list[Linkable]
    url_index_cache: dict[str, int]

    def __init__(self):
        """Init a new registry."""
        self.links = []

    def add_linkable(
        self, url: str, data_type: str, idx: str, _refresh_cache: bool = True
    ) -> None:
        """Add a single linkable item to the registry.

        The URL will be compared later against a list of links that
        need to be resolved and the data type and IDX will be returned.

        Data types should be unique.
        IDXes should be unique within one or more data types.

        Args:
            url: The URL of the destination
            data_type: A unique identifier for this type of item.
            idx: A unique identifier within the data type.
            _refresh_cache: Whether the link cache should be updated. Should be left as
                True unless multiple links are being added together.
        """
        self.links.append(Linkable(link=url, data_type=data_type, idx=idx))

        if _refresh_cache:
            self._refresh_cache()

    def add_linkables(self, data_type: str, links: list[str], idxes: list[str]) -> None:
        """Add multiple linkable items at once.

        Args:
            data_type: The data type for all items.
            links: A list of links. Must be the same length as idxes.
            idxes: A list of IDs. Must be the same length as links.

        Raises:
            ValueError: if the links and idxes lists are not the same length.

        """
        if len(links) != len(idxes):
            raise ValueError(
                f"Links and idxes must be same length ({len(links)} != {len(idxes)})"
            )

        for link, idx in zip(links, idxes):
            self.add_linkable(
                data_type=data_type, url=link, idx=idx, _refresh_cache=False
            )
        self._refresh_cache()

    def _refresh_cache(self):
        self.url_index_cache = {}

        for i, link in enumerate(self.links):
            self.url_index_cache[link.link] = i

    def query_link(self, href: str) -> Optional[Linkable]:
        """Find a linkable item by the URL in the registry.

        Returns None if no URL matches.

        Args:
            href: A URL to search

        Returns:
            A matching linkable
        """
        if href in self.url_index_cache:
            return self.links[self.url_index_cache[href]]
        else:
            return None



import platform

import click

from wpextract.cli._download import download
from wpextract.cli._extract import extract
from wpextract.cli._shared import CMD_ARGS

PYTHON_VERSION = platform.python_version()


@click.group(context_settings=dict(help_option_names=["-h", "--help"]), **CMD_ARGS)
@click.version_option(
    package_name="wpextract",
    message=f"%(prog)s, version %(version)s (Python {PYTHON_VERSION})",
)
def cli():
    """WPextract is a tool to create datasets from WordPress sites."""


cli.add_command(download)
cli.add_command(extract)



from pathlib import Path
from typing import Optional

import click

from wpextract import WPExtractor
from wpextract.cli._shared import (
    CMD_ARGS,
    directory,
    empty_directory,
    logging_options,
    setup_logging,
    setup_tqdm_redirect,
)


@click.command(short_help="Extract site to a dataset.", **CMD_ARGS)
@click.argument("json_root", type=directory)
@click.argument(
    "out_dir", type=click.Path(), callback=empty_directory, metavar="DIRECTORY"
)
@click.option(
    "-S", "--scrape-root", help="Root directory of an HTML scrape", type=directory
)
@click.option(
    "-P", "--json-prefix", help="Prefix to use for input and output filenames"
)
@logging_options
def extract(
    json_root: Path,
    out_dir: Path,
    scrape_root: Optional[Path],
    json_prefix: Optional[str],
    log: Optional[Path],
    verbose: bool,
):
    """Converts the downloaded data files into a dataset.

    JSON_ROOT is a directory containing a JSON dump of the data files, such as one generated with wpextract download.

    OUT_DIR is the directory to output the extracted JSON to. It must be an existing empty directory or a non-existent directory which will be created.
    """
    setup_logging(verbose, log)

    with setup_tqdm_redirect(log is None):
        extractor = WPExtractor(
            json_root=json_root,
            scrape_root=scrape_root,
            json_prefix=json_prefix,
        )
        extractor.extract()
        extractor.export(out_dir)



from pathlib import Path
from typing import Optional

import pandas as pd

from wpextract.extractors.io import load_df

EXPORT_COLUMNS = ["avatar", "description", "link", "name", "slug", "url"]


def load_users(path: Path) -> Optional[pd.DataFrame]:
    """Load the users from a JSON file.

    Args:
        path: Path to file.

    Returns:
        Dataframe of users
    """
    users_df = load_df(path)

    if users_df is None:
        return None

    if "yoast_head_json.og_image" in users_df.columns:
        users_df["avatar"] = users_df["yoast_head_json.og_image"].apply(
            lambda image: image[0]["url"] if len(image) > 0 else None
        )
    else:
        users_df["avatar"] = None

    users_df = users_df[users_df.columns.intersection(EXPORT_COLUMNS)]

    return users_df



import copy
import html
import json
from pathlib import Path
from urllib import parse as urlparse

from tqdm.auto import tqdm

from wpextract.download.requestsession import RequestSession


class Exporter:
    """Utility functions to export data."""

    CHUNK_SIZE = 2048
    """The size of chunks to download large files"""

    @staticmethod
    def download_media(
        session: RequestSession, media: list[str], out_path: Path
    ) -> int:
        """Downloads the media files based on the given URLs.

        Args:
            session: the request session to use
            media: the URLs as a list
            out_path: the path to the folder where the files are being saved, it is assumed as existing

        Returns:
            the number of files written
        """
        files_number = 0
        for m in tqdm(media, unit="media"):
            r = session.do_request("get", m, stream=True)
            if r.status_code == 200:
                http_path = urlparse.urlparse(m).path.split("/")
                local_path = out_path
                if len(http_path) > 1:
                    for el in http_path[:-1]:
                        local_path = local_path / el
                        local_path.mkdir(exist_ok=True)
                local_path = local_path / http_path[-1]
                with open(local_path, "wb") as f:
                    i = 0
                    content_size = int(r.headers.get("Content-Length", -1))
                    chunks_pbar = None
                    if content_size > 0:
                        chunks_pbar = tqdm(
                            total=content_size,
                            unit="B",
                            unit_scale=True,
                            desc=http_path[-1],
                            leave=False,
                        )
                    for chunk in r.iter_content(Exporter.CHUNK_SIZE):
                        if chunks_pbar is not None:
                            chunks_pbar.update(Exporter.CHUNK_SIZE)
                        f.write(chunk)
                        i += 1
                    if chunks_pbar is not None:
                        chunks_pbar.update(content_size - chunks_pbar.n)
                        chunks_pbar.close()
                files_number += 1
        return files_number

    @staticmethod
    def setup_export(vlist, parameters_to_unescape):
        """Sets up the right values for a list export.

        This function flattens alist of objects before its serialization in the expected format.
        It also makes a deepcopy to ensure that the original vlist is not altered.

        Args:
            vlist: the list to prepare for exporting
            parameters_to_unescape: parameters to unescape (ex.
                ["param1", ["param2"]["rendered"]])
        """
        exported_list = []

        for el in vlist:
            if el is not None:
                # First copy the object
                exported_el = copy.deepcopy(el)
                # Look for parameters to HTML unescape
                for key in parameters_to_unescape:
                    if type(key) is str:  # If the parameter is at the root
                        exported_el[key] = html.unescape(exported_el[key])
                    elif type(key) is list:  # If the parameter is nested
                        selected = exported_el
                        siblings = []
                        fullpath = {}
                        # We look for the leaf first, not forgetting sibling branches for rebuilding the tree later
                        for k in key:
                            if type(selected) is dict and k in selected.keys():
                                sib = {}
                                for e in selected.keys():
                                    if e != k:
                                        sib[e] = selected[e]
                                selected = selected[k]
                                siblings.append(sib)
                            else:
                                selected = None
                                break
                        # If we can unescape the parameter, we do it and rebuild the tree starting from the leaf
                        if selected is not None and type(selected) is str:
                            selected = html.unescape(selected)
                            key.reverse()
                            fullpath[key[0]] = selected
                            s = len(siblings) - 1
                            for e in siblings[s].keys():
                                fullpath[e] = siblings[s][e]
                            for k in key[1:]:
                                fullpath = {k: fullpath}
                                s -= 1
                                for e in siblings[s].keys():
                                    fullpath[e] = siblings[s][e]
                            key.reverse()
                            exported_el[key[0]] = fullpath[key[0]]
                # The resulting element is appended to the list of exported elements
                exported_list.append(exported_el)

        return exported_list

    @staticmethod
    def write_file(filename, data):
        """Writes content to the given file in JSON format.

        The key mapping must be a dict of keys or lists of keys to ensure proper mapping.

        Args:
            filename: the path of the file
            data: the actual data to export
        """
        with open(filename, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)

    @staticmethod
    def export_posts(
        posts: list[dict],
        filename: str,
    ):
        """Exports posts to the specified file.

        Args:
            posts: the posts to export
            filename: filename to use

        Returns:
            the length of the list written to the file
        """
        exported_posts = Exporter.setup_export(
            posts,
            [["title", "rendered"], ["content", "rendered"], ["excerpt", "rendered"]],
        )

        Exporter.write_file(filename, exported_posts)
        return len(exported_posts)

    @staticmethod
    def export_categories(categories, filename):
        """Exports categories to the specified file.

        Args:
            categories: the categories to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_categories = Exporter.setup_export(
            categories,  # TODO
            [],
        )

        Exporter.write_file(filename, exported_categories)
        return len(exported_categories)

    @staticmethod
    def export_tags(tags, filename):
        """Exports tags to the specified file.

        Args:
            tags: the tags to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_tags = tags  # It seems that no modification will be done for this one, so no deepcopy
        Exporter.write_file(filename, exported_tags)
        return len(exported_tags)

    @staticmethod
    def export_users(users, filename):
        """Exports users to the specified file.

        Args:
            users: the users to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_users = users  # It seems that no modification will be done for this one, so no deepcopy
        Exporter.write_file(filename, exported_users)
        return len(exported_users)

    @staticmethod
    def export_pages(pages, filename):
        """Exports pages to the specified file.

        Args:
            pages: the pages to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_pages = Exporter.setup_export(
            pages,
            [
                ["guid", "rendered"],
                ["title", "rendered"],
                ["content", "rendered"],
                ["excerpt", "rendered"],
            ],
        )

        Exporter.write_file(filename, exported_pages)
        return len(exported_pages)

    @staticmethod
    def export_media(media, filename):
        """Exports media to the specified file.

        Args:
            media: the media to export
            filename: file to export to

        Returns:
            the length of the list written to the file
        """
        exported_media = Exporter.setup_export(
            media,
            [
                ["guid", "rendered"],
                ["title", "rendered"],
                ["description", "rendered"],
                ["caption", "rendered"],
            ],
        )

        Exporter.write_file(filename, exported_media)
        return len(exported_media)

    # FIXME to be refactored
    @staticmethod
    def export_comments_interactive(comments, filename):
        """Exports comments to the specified file.

        Args:
            comments: the comments to export
            filename: the path to the file to write

        Returns:
            the length of the list written to the file
        """
        exported_comments = Exporter.setup_export(
            comments,
            [["content", "rendered"]],
        )

        Exporter.write_file(filename, exported_comments)
        return len(exported_comments)



from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd

from wpextract.extractors.data.links import LinkRegistry
from wpextract.extractors.io import load_df
from wpextract.util.locale import extract_locale

EXPORT_COLUMNS = [
    "name",
    "slug",
    "description",
    "count",
    "link",
    "link_locale",
    "parent",
]


def load_categories(path: Path, link_registry: LinkRegistry) -> Optional[pd.DataFrame]:
    """Load the categories from a JSON file.

    The JSON file is expected to be in the response format of the WordPress
    categories API.

    Args:
        path: The path of the JSON file
        link_registry: A link registry to populate

    Returns:
        A dataframe of the categories.
    """
    categories_df = load_df(path)

    if categories_df is None:
        return None

    categories_df["parent"] = categories_df["parent"].replace(0, np.nan)
    categories_df["link_locale"] = categories_df["link"].apply(extract_locale)

    categories_df = categories_df[categories_df.columns.intersection(EXPORT_COLUMNS)]

    link_registry.add_linkables(
        "category", categories_df["link"].to_list(), categories_df.index.to_list()
    )

    return categories_df



import logging
import random
import time
from http.cookies import SimpleCookie
from typing import Optional, Union

import requests
from requests.adapters import HTTPAdapter
from requests.auth import HTTPBasicAuth, HTTPDigestAuth
from urllib3 import Retry

DEFAULT_UA = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36"


class ConnectionCouldNotResolve(Exception):
    """The remote host could not be resolved."""

    pass


class ConnectionReset(Exception):
    """The connection was reset during the request."""

    pass


class ConnectionRefused(Exception):
    """The connection was refused by the server."""

    pass


class ConnectionTimeout(Exception):
    """A connection timeout occurred."""

    pass


class HTTPError400(Exception):
    """HTTP Bad Request.

    See Also:
        HTTPErrorInvalidPage for a special case of this error.
    """

    pass


class HTTPErrorInvalidPage(Exception):
    """Special case of HTTP 400 if the error is caused by a nonexistent page.

    This indicates the last page has been passed and all items have been retrieved.
    """

    pass


class HTTPError401(Exception):
    """HTTP Unauthorized."""

    pass


class HTTPError403(Exception):
    """HTTP Forbidden."""

    pass


class HTTPError404(Exception):
    """HTTP Not Found."""

    pass


class HTTPError500(Exception):
    """HTTP Internal Server Error."""

    pass


class HTTPError502(Exception):
    """HTTP Bad Gateway."""

    pass


class HTTPError(Exception):
    """A generic HTTP error with an unexpected code."""

    pass


class HTTPTooManyRedirects(Exception):
    """Raised if the number of allowed redirects exceeds the configured maximum value."""

    pass


RETRY_AFTER_STATUS = frozenset(
    [
        # Standard HTTP
        413,  # Payload Too Large
        429,  # Too Many Requests
        500,  # Internal Server Error
        502,  # Bad Gateway
        503,  # Service Unavailable
        504,  # Gateway Timeout,
        # Cloudflare
        520,  # Web Server Returned an Unknown Error
        522,  # Connection Timed Out
        524,  # A Timeout Occurred
    ]
)

ERROR_NAMES = {
    400: "Bad Request",
    401: "Unauthorized",
    403: "Forbidden",
    404: "Not Found",
    500: "Internal Server Error",
    502: "Bad Gateway",
}
EXCEPTION_CLS = {
    400: HTTPError400,
    401: HTTPError401,
    403: HTTPError403,
    404: HTTPError404,
    500: HTTPError500,
    502: HTTPError502,
}


def _handle_status(url, status_code, n_tries=None):
    if 300 <= status_code < 400:
        logging.error(
            f'Too many redirects (status code {status_code}) while fetching "{url}"'
        )
        raise HTTPTooManyRedirects

    if status_code in ERROR_NAMES:
        log_msg = (
            f'Error {status_code} ({ERROR_NAMES[status_code]}) while fetching "{url}"'
        )
    else:
        log_msg = f'Error {status_code} while fetching "{url}"'

    if n_tries is not None:
        log_msg += f" after {n_tries} tries"

    if status_code in EXCEPTION_CLS or status_code > 400:
        if status_code == 404:
            logging.debug(log_msg)
        else:
            logging.error(log_msg)

        if status_code in EXCEPTION_CLS:
            raise EXCEPTION_CLS[status_code]
        else:
            raise HTTPError


class RequestWait:
    """Manages waiting between requests."""

    def __init__(
        self, wait: Optional[float] = None, random_wait: Optional[bool] = False
    ):
        """Create a new waiting instance.

        Args:
            wait: The base time to wait in seconds
            random_wait: If true, the wait duration will be multiplied by a random factor between 0.5 and 1.5.
        """
        self.wait_s = wait or 0
        self.random_wait = random_wait

    def wait(self):
        """Perform the specified wait."""
        if self.wait is None:
            return

        wait_factor = 1
        if self.random_wait:
            wait_factor = random.uniform(0.5, 1.5)

        time.sleep(self.wait_s * wait_factor)


AuthorizationType = Union[tuple[str, str], HTTPBasicAuth, HTTPDigestAuth]


class RequestSession:
    """Manages HTTP requests and their behaviour."""

    def __init__(
        self,
        proxy: Optional[str] = None,
        cookies: Optional[str] = None,
        authorization: Optional[AuthorizationType] = None,
        timeout: Optional[float] = 30,
        wait: Optional[float] = None,
        random_wait: bool = False,
        max_retries: int = 10,
        backoff_factor: float = 0.1,
        max_redirects: int = 20,
    ):
        """Create a new request session.

        Args:
            proxy: a dict containing a proxy server string for HTTP and/or HTTPS connection
            cookies: a string in the format of the Cookie header
            authorization: a tuple containing login and password or [`requests.auth.HTTPBasicAuth`][requests.auth.HTTPBasicAuth] for basic authentication or [`requests.auth.HTTPDigestAuth`][requests.auth.HTTPDigestAuth] for NTLM-like authentication
            timeout: maximum time in seconds to wait for a response before giving up
            wait: wait time in seconds between requests, None to not wait
            random_wait: If true, the wait time between requests is multiplied by a random factor between 0.5 and 1.5
            max_retries: the maximum number of retries before failing
            backoff_factor: Factor to wait between successive retries
            max_redirects: maximum number of redirects to follow
        """
        self.s = requests.Session()
        if proxy is not None:
            self.set_proxy(proxy)
        if cookies is not None:
            self.set_cookies(cookies)
        if authorization is not None and (
            type(authorization) is tuple
            and len(authorization) == 2
            or type(authorization) is HTTPBasicAuth
            or type(authorization) is HTTPDigestAuth
        ):
            self.s.auth = authorization
        self.wait = wait
        self.timeout = timeout
        self._mount_retry(backoff_factor, max_redirects, max_retries)
        self.waiter = RequestWait(wait, random_wait)

    def _mount_retry(self, backoff_factor, max_redirects, max_retries):
        retry = Retry(
            total=max_retries,
            backoff_factor=backoff_factor,
            redirect=max_redirects,
            status_forcelist=RETRY_AFTER_STATUS,
            raise_on_redirect=False,
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.s.mount("http://", adapter)
        self.s.mount("https://", adapter)

    def get(self, url):
        """Calls the get function from requests but handles errors to raise proper exception following the context."""
        return self.do_request("get", url)

    def post(self, url, data=None):
        """Calls the post function from requests but handles errors to raise proper exception following the context."""
        return self.do_request("post", url, data)

    def do_request(self, method, url, data=None, stream=False):
        """Helper class to regroup requests and handle exceptions at the same location."""
        headers = {"User-Agent": DEFAULT_UA}
        response = None
        try:
            if method == "post":
                response = self.s.post(url, data, headers=headers, timeout=self.timeout)
            else:
                response = self.s.get(
                    url, headers=headers, timeout=self.timeout, stream=stream
                )
        except requests.ConnectionError as e:
            if "Errno -5" in str(e) or "Errno -2" in str(e) or "Errno -3" in str(e):
                logging.error(f"Could not resolve host {url}")
                raise ConnectionCouldNotResolve from e
            elif "Errno 111" in str(e):
                logging.error(f"Connection refused by {url}")
                raise ConnectionRefused from e
            elif "RemoteDisconnected" in str(e):
                logging.error(f"Connection reset by {url}")
                raise ConnectionReset from e
            else:
                raise e
        except requests.Timeout as e:
            logging.error(f"Request timed out fetching {url}")
            raise ConnectionTimeout from e

        # If this is an HTTP 400 due to an invalid page, raise this special error early
        if response.status_code == 400 and "application/json" in response.headers.get(
            "content-type", ""
        ):
            json_body = response.json()
            if "code" in json_body and "invalid_page_number" in json_body["code"]:
                logging.debug(
                    "Received HTTP 400 error which appears to be an invalid page error, probably reached the end."
                )
                raise HTTPErrorInvalidPage

        n_tries = None
        if hasattr(response.raw, "retries") and response.raw.retries is not None:
            # initial request + n retires = n+1 tries
            n_tries = len(response.raw.retries.history) + 1

        _handle_status(url, response.status_code, n_tries)

        self.waiter.wait()
        return response

    def set_cookies(self, cookies: str) -> None:
        """Sets new cookies from a string.

        Args:
            cookies: Cookies in a format usable by [http.cookies.SimpleCookie.load][]
        """
        c = SimpleCookie()
        c.load(cookies)
        for key, m in c.items():
            self.s.cookies.set(key, m.value)

    def get_cookies(self) -> dict[str, str]:
        """Retrieve cookies in the cookie jar.

        Returns:
            Cookies in dictionary format.
        """
        return self.s.cookies.get_dict()

    def set_proxy(self, proxy) -> None:
        """Set a proxy to use for the request.

        Args:
            proxy: proxy URL In the format supported by requests
        """
        prot = "http"
        if proxy[:5].lower() == "https":
            prot = "https"
        self.s.proxies = {prot: proxy}

    def set_creds(self, credentials: AuthorizationType) -> None:
        """Set new credentials for the request.

        Args:
            credentials: credentials as an HTTPBasic tuple or supported requests HTTP auth instance
        """
        self.s.auth = credentials



from pathlib import Path
from typing import Optional

import pandas as pd
from bs4 import Tag

from wpextract.extractors.data.links import LinkRegistry
from wpextract.extractors.io import load_df
from wpextract.parse.html import extract_html_text

EXPORT_COLUMNS = [
    "alt_text",
    "author",
    "caption.rendered",
    "caption.text",
    "date_gmt",
    "description.rendered",
    "description.text",
    "guid.rendered",
    "media_details.bitrate",
    "media_details.file",
    "media_details.image_meta.camera",
    "media_details.image_meta.created_timestamp",
    "media_details.image_meta.credit",
    "media_details.image_meta.focal_length",
    "media_details.image_meta.iso",
    "media_details.image_meta.orientation",
    "media_details.length",
    "media_details.mime_type",
    "media_details.original_image",
    "media_details.parent_image.attachment_id",
    "media_type",
    "mime_type",
    "modified_gmt",
    "post",
    "slug",
    "source_url",
    "title.rendered",
    "title.text",
    "yoast_head_json.og_url",
    "yoast_head_json.title",
]

RENAME_COLUMNS = {
    "caption.rendered": "caption.html",
    "description.rendered": "description.html",
    "guid.rendered": "guid",
    "post": "post_id",
    "title.rendered": "title.html",
    "yoast_head_json.title": "page_title",
    "media_details.parent_image.attachment_id": "parent_image_id",
    "yoast_head_json.og_url": "og_url",
}


def load_media(path: Path, link_registry: LinkRegistry) -> Optional[pd.DataFrame]:
    """Load media from a JSON file.

    The JSON file is expected to be in the response format of the WordPress media API.

    Args:
        path: The path to the JSON file
        link_registry: A link registry to populate

    Returns:
        A dataframe of the media
    """
    media_df = load_df(path)

    if media_df is None:
        return None

    # Convert times
    media_df["date_gmt"] = pd.to_datetime(media_df["date_gmt"])
    media_df["modified_gmt"] = pd.to_datetime(media_df["modified_gmt"])
    media_df = media_df.drop(["date", "modified"], axis=1)

    # Convert post ID to a NA-able integer
    media_df["post"] = media_df["post"].astype("Int64")

    media_df["description.text"] = media_df["description.rendered"].apply(
        extract_html_text
    )
    media_df["caption.text"] = media_df["caption.rendered"].apply(extract_html_text)
    media_df["title.text"] = media_df["title.rendered"].apply(extract_html_text)

    media_df.loc[media_df["description.text"] == "\n", "description.text"] = ""

    media_df = media_df[media_df.columns.intersection(EXPORT_COLUMNS)]

    media_df = media_df.rename(columns=RENAME_COLUMNS, errors="ignore")

    link_registry.add_linkables(
        "media", media_df["source_url"].to_list(), media_df.index.to_list()
    )

    return media_df


def get_caption(img: Tag) -> Optional[str]:
    """Get the caption of an image tag.

    Searches adjacent tags for a <figcaption>.

    TODO: A heuristic to find likely non-figcaption captions.

    Raises:
        ValueError: if a non-image tag is passed.

    Args:
        img: An img tag

    Returns:
        The caption text or None if there is no caption.
    """
    if img.name != "img":
        raise ValueError("Attempting to get caption of non-image")

    parent_caption = img.find_parent("figure")

    if parent_caption is None:
        return None

    caption = parent_caption.find("figcaption")

    if caption is None:
        return None

    return caption.get_text()



import functools
import logging
import pathlib
from contextlib import contextmanager, nullcontext
from typing import Callable, Optional

import click
from click import BadParameter
from click_option_group import optgroup
from tqdm.contrib.logging import logging_redirect_tqdm

CMD_ARGS = {
    "epilog": "See https://gatenlp.github.io/wordpress-site-extractor/ for documentation.",
}


file = click.Path(
    exists=False, dir_okay=False, file_okay=True, writable=True, path_type=pathlib.Path
)
directory = click.Path(
    exists=True,
    file_okay=False,
    dir_okay=True,
    readable=True,
    writable=False,
    path_type=pathlib.Path,
)


def empty_directory(ctx, param, value: str):
    if value is None:
        return value

    path = pathlib.Path(value)

    if path.exists() and not path.is_dir():
        raise BadParameter("exists but is not a directory")

    try:
        path.mkdir(exist_ok=True)
    except OSError as e:
        raise BadParameter("directory could not be created") from e

    if any(path.iterdir()):
        raise BadParameter("is not empty, must be an empty directory")

    return path


def logging_options(cmd_func: Callable) -> Callable:
    @optgroup.group("logging")
    @optgroup.option(
        "-l", "--log", type=file, help="File to log to, will suppress stdout."
    )
    @optgroup.option(
        "-v", "--verbose", is_flag=True, help="Increase log level to include debug logs"
    )
    @functools.wraps(cmd_func)
    def wrapper(*args, **kwargs):
        return cmd_func(*args, **kwargs)

    return wrapper


def setup_logging(verbose: bool, log_path: Optional[pathlib.Path]) -> None:
    log_level = logging.DEBUG if verbose else logging.INFO

    if log_path is not None:
        logging.basicConfig(filename=log_path, level=log_level)
    else:
        logging.basicConfig(level=log_level)


@contextmanager
def setup_tqdm_redirect(should_redirect: bool):
    """Conditionally yields the tqdm log redirect context.

    If `should_redirect` is false, the [`nullcontext`][contextlib.nullcontext] is yielded.
    This is a valid context but has no effect, allowing this function to always be used in a with.

    This is necessary because, contrary to what the docs imply, using logging_redirect_tqdm
    when a log file is set causes logs to be output to stdout.

    Args:
        should_redirect: whether to apply the redirect.

    Yields:
        tqdm log redirect context or null context
    """
    with logging_redirect_tqdm() if should_redirect else nullcontext():
        yield



import copy
import logging
import math
from json.decoder import JSONDecodeError
from urllib.parse import urlencode

from tqdm.auto import tqdm

from wpextract.download.exceptions import (
    NoWordpressApi,
    NSNotFoundException,
    WordPressApiNotV2,
)
from wpextract.download.requestsession import (
    HTTPError404,
    HTTPErrorInvalidPage,
    RequestSession,
)
from wpextract.download.utils import (
    get_by_id,
    get_content_as_json,
    url_path_join,
)


class WPApi:
    """Queries the WordPress API to retrieve information."""

    # Object types
    POST = 0
    """The post type"""
    POST_REVISION = 1
    """The post revision type"""
    WP_BLOCK = 2
    """The Gutenberg block type"""
    CATEGORY = 3
    """The category type"""
    TAG = 4
    """The tag type"""
    PAGE = 5
    """The page type"""
    COMMENT = 6
    """The comment type"""
    MEDIA = 7
    """The media type"""
    USER = 8
    """The user type"""
    THEME = 9
    """The theme type"""
    NAMESPACE = 10
    """The namespace type"""
    # SEARCH_RESULT = 10
    ALL_TYPES = 20
    """Constant representing all types"""

    def __init__(self, target, api_path="wp-json/", session=None, search_terms=None):
        """Creates a new instance of WPApi.

        Args:
            target: the target of the scan
            api_path: the api path, if non-default
            session: the requests session object to use for HTTP requests
            search_terms : the terms of the keyword search, if any
        """
        self.api_path = api_path
        self.search_terms = search_terms
        self.has_v2 = None
        self.name = None
        self.description = None
        self.url = target
        self.basic_info = None
        self.posts = None
        self.tags = None
        self.categories = None
        self.users = None
        self.media = None
        self.pages = None
        self.s = None
        self.comments_loaded = False
        self.orphan_comments = []
        self.comments = None

        if session is not None:
            self.s = session
        else:
            self.s = RequestSession()

    def get_orphans_comments(self):
        """Returns the list of comments for which a post hasn't been found."""
        return self.orphan_comments

    def get_basic_info(self):
        """Collects and stores basic information about the target."""
        rest_url = url_path_join(self.url, self.api_path)
        if self.basic_info is not None:
            return self.basic_info

        try:
            req = self.s.get(rest_url)
        except Exception as e:
            raise NoWordpressApi from e
        if req.status_code >= 400:
            raise NoWordpressApi
        self.basic_info = get_content_as_json(req)

        if "name" in self.basic_info.keys():
            self.name = self.basic_info["name"]

        if "description" in self.basic_info.keys():
            self.description = self.basic_info["description"]

        if (
            "namespaces" in self.basic_info.keys()
            and "wp/v2" in self.basic_info["namespaces"]
        ):
            self.has_v2 = True

        return self.basic_info

    def crawl_pages(
        self, url, start=None, num=None, search_terms=None, display_progress=True
    ):
        """Crawls all pages while there is at least one result for the given endpoint or tries to get pages from start to end."""
        if search_terms is None:
            search_terms = self.search_terms
        page = 1
        total_entries = 0
        total_pages = 0
        more_entries = True
        entries = []
        base_url = url
        entries_left = 1
        per_page = 10
        if search_terms is not None:
            if "?" in base_url:
                base_url += "&" + urlencode({"search": search_terms})
            else:
                base_url += "?" + urlencode({"search": search_terms})
        if start is not None:
            page = math.floor(start / per_page) + 1
        if num is not None:
            entries_left = num

        # Initialise placeholder for progress bar
        pbar = None

        while more_entries and entries_left > 0:
            rest_url = url_path_join(self.url, self.api_path, (base_url % page))
            if start is not None:
                rest_url += "&per_page=%d" % per_page
            try:
                req = self.s.get(rest_url)
                if (
                    page == 1
                    or start is not None
                    and page == math.floor(start / per_page) + 1
                ) and "X-WP-Total" in req.headers:
                    total_entries = int(req.headers["X-WP-Total"])
                    total_pages = int(req.headers["X-WP-TotalPages"])
                    logging.info("Total number of entries: %d" % total_entries)
                    if start is not None and total_entries < start:
                        start = total_entries - 1
            except HTTPErrorInvalidPage:
                break
            except Exception as e:
                raise WordPressApiNotV2 from e

            try:
                json_content = get_content_as_json(req)
                if type(json_content) is list and len(json_content) > 0:
                    if (
                        start is None
                        or start is not None
                        and page > math.floor(start / per_page) + 1
                    ) and num is None:
                        entries += json_content
                        if start is not None:
                            entries_left -= len(json_content)
                    elif start is not None and page == math.floor(start / per_page) + 1:
                        if (
                            num is None
                            or num is not None
                            and len(json_content[start % per_page :]) < num
                        ):
                            entries += json_content[start % per_page :]
                            if num is not None:
                                entries_left -= len(json_content[start % per_page :])
                        else:
                            entries += json_content[
                                start % per_page : (start % per_page) + num
                            ]
                            entries_left = 0
                    else:
                        if num is not None and entries_left > len(json_content):
                            entries += json_content
                            entries_left -= len(json_content)
                        else:
                            entries += json_content[:entries_left]
                            entries_left = 0

                    if display_progress:
                        if num is None and start is None and total_entries >= 0:
                            if pbar is None:
                                pbar = tqdm(total=total_pages)
                            pbar.update(page - pbar.n)

                        elif num is None and start is not None and total_entries >= 0:
                            if pbar is None:
                                pbar = tqdm(total=total_entries - start)
                            pbar.update(total_entries - start - entries_left - pbar.n)
                        elif num is not None and total_entries > 0:
                            if pbar is None:
                                pbar = tqdm(total=total_entries)
                            pbar.update(num - entries_left - pbar.n)
                else:
                    more_entries = False
            except JSONDecodeError:
                more_entries = False

            page += 1

        if pbar is not None:
            pbar.close()

        return (entries, total_entries)

    def crawl_single_page(self, url):
        """Crawls a single URL."""
        content = None
        rest_url = url_path_join(self.url, self.api_path, url)
        try:
            req = self.s.get(rest_url)
        except HTTPErrorInvalidPage:
            return None
        except HTTPError404:
            return None
        except Exception as e:
            raise WordPressApiNotV2 from e
        try:
            content = get_content_as_json(req)
        except JSONDecodeError:
            pass

        return content

    def get_from_cache(self, cache, start=None, num=None, force=False):
        """Tries to fetch data from the given cache, also verifies first if WP-JSON is supported."""
        if self.has_v2 is None:
            self.get_basic_info()
        if not self.has_v2:
            raise WordPressApiNotV2
        if cache is not None and start is not None and len(cache) <= start:
            start = len(cache) - 1
        if cache is not None and not force:
            if (
                start is not None
                and num is None
                and len(cache) > start
                and None not in cache[start:]
            ):
                # If start is specified and not num, we want to return the posts in cache only if they were already cached
                return cache[start:]
            elif (
                start is None
                and num is not None
                and len(cache) > num
                and None not in cache[:num]
            ):
                # If num is specified and not start, we want to do something similar to the above
                return cache[:num]
            elif (
                start is not None
                and num is not None
                and len(cache) > start + num
                and None not in cache[start:num]
            ):
                return cache[start : start + num]
            elif (
                start is None and (num is None or num > len(cache))
            ) and None not in cache:
                return cache

        return None

    def update_cache(self, cache, values, total_entries, start=None, num=None):
        """Push new values to the cache."""
        if cache is None:
            cache = values
        elif len(values) > 0:
            s = start
            if start is None:
                s = 0
            if start >= total_entries:
                s = total_entries - 1
            n = num
            if n is not None and s + n > total_entries:
                n = total_entries - s
            if num is None:
                n = total_entries
            if n > len(cache):
                cache += [None] * (n - len(cache))
            for el in values:
                cache[s] = el
                s += 1
                if s == n:
                    break
        if len(cache) != total_entries:
            if start is not None and start < total_entries:
                cache = [None] * start + cache
            if num is not None:
                cache += [None] * (total_entries - len(cache))
        return cache

    def get_comments(self, start=None, num=None, force=False):
        """Retrieves all comments."""
        comments = self.get_from_cache(self.comments, start, num, force)
        if comments is not None:
            return comments

        comments, total_entries = self.crawl_pages("wp/v2/comments?page=%d", start, num)
        self.comments = self.update_cache(
            self.comments, comments, total_entries, start, num
        )
        return comments

    def get_posts(self, comments=False, start=None, num=None, force=False):
        """Retrieves all posts or the specified ones."""
        if self.has_v2 is None:
            self.get_basic_info()
        if not self.has_v2:
            raise WordPressApiNotV2
        if self.posts is not None and start is not None and len(self.posts) < start:
            start = len(self.posts) - 1
        if (
            self.posts is not None
            and (self.comments_loaded and comments or not comments)
            and not force
        ):
            posts = self.get_from_cache(self.posts, start, num)
            if posts is not None:
                return posts
        posts, total_entries = self.crawl_pages(
            "wp/v2/posts?page=%d", start=start, num=num
        )

        self.posts = self.update_cache(self.posts, posts, total_entries, start, num)

        if not self.comments_loaded and comments:
            # Load comments
            comment_list = self.crawl_pages("wp/v2/comments?page=%d")[0]
            for comment in comment_list:
                found_post = False
                for i in range(0, len(self.posts)):
                    if self.posts[i]["id"] == comment["post"]:
                        if "comments" not in self.posts[i]:
                            self.posts[i]["comments"] = []
                        self.posts[i]["comments"].append(comment)
                        found_post = True
                        break
                if not found_post:
                    self.orphan_comments.append(comment)
            self.comments_loaded = True

        return_posts = self.posts
        if start is not None and start < len(return_posts):
            return_posts = return_posts[start:]
        if num is not None and num < len(return_posts):
            return_posts = return_posts[:num]
        return return_posts

    def get_tags(self, start=None, num=None, force=False):
        """Retrieves all tags."""
        tags = self.get_from_cache(self.tags, start, num, force)
        if tags is not None:
            return tags

        tags, total_entries = self.crawl_pages("wp/v2/tags?page=%d", start, num)
        self.tags = self.update_cache(self.tags, tags, total_entries, start, num)
        return tags

    def get_categories(self, start=None, num=None, force=False):
        """Retrieves all categories or the specified ones."""
        categories = self.get_from_cache(self.categories, start, num, force)
        if categories is not None:
            return categories

        categories, total_entries = self.crawl_pages(
            "wp/v2/categories?page=%d", start=start, num=num
        )
        self.categories = self.update_cache(
            self.categories, categories, total_entries, start, num
        )
        return categories

    def get_users(self, start=None, num=None, force=False):
        """Retrieves all users or the specified ones."""
        users = self.get_from_cache(self.users, start, num, force)
        if users is not None:
            return users

        users, total_entries = self.crawl_pages(
            "wp/v2/users?page=%d", start=start, num=num
        )
        self.users = self.update_cache(self.users, users, total_entries, start, num)
        return users

    def get_media(self, start=None, num=None, force=False):
        """Retrieves all media objects."""
        media = self.get_from_cache(self.media, start, num, force)
        if media is not None:
            return media

        media, total_entries = self.crawl_pages(
            "wp/v2/media?page=%d", start=start, num=num
        )
        self.media = self.update_cache(self.media, media, total_entries, start, num)
        return media

    def get_media_urls(self, ids, cache=True):
        """Retrieves the media download URLs for specified IDs or all or from cache."""
        media = []
        if ids == "all":
            media = self.get_media(force=(not cache))
        elif ids == "cache":
            media = self.get_from_cache(self.media, force=(not cache))
        else:
            id_list = ids.split(",")
            media = []
            for i in id_list:
                try:
                    if int(i) > 0:
                        m = self.get_obj_by_id(WPApi.MEDIA, int(i), cache)
                        if m is not None and len(m) > 0 and type(m[0]) is dict:
                            media.append(m[0])
                except ValueError:
                    pass
        urls = []
        slugs = []
        if media is None:
            return []
        for m in media:
            if (
                m is not None
                and type(m) is dict
                and "source_url" in m.keys()
                and "slug" in m.keys()
            ):
                urls.append(m["source_url"])
                slugs.append(m["slug"])
        return urls, slugs

    def get_pages(self, start=None, num=None, force=False):
        """Retrieves all pages."""
        pages = self.get_from_cache(self.pages, start, num, force)
        if pages is not None:
            return pages

        pages, total_entries = self.crawl_pages(
            "wp/v2/pages?page=%d", start=start, num=num
        )
        self.pages = self.update_cache(self.pages, pages, total_entries, start, num)
        return pages

    def get_namespaces(self, start=None, num=None, force=False):
        """Retrieves an array of namespaces."""
        if self.has_v2 is None or force:
            self.get_basic_info()
        if "namespaces" in self.basic_info.keys():
            if start is None and num is None:
                return self.basic_info["namespaces"]
            namespaces = copy.deepcopy(self.basic_info["namespaces"])
            if start is not None and start < len(namespaces):
                namespaces = namespaces[start:]
            if num <= len(namespaces):
                namespaces = namespaces[:num]
            return namespaces
        return []

    def get_routes(self):
        """Retrieves an array of routes."""
        if self.has_v2 is None:
            self.get_basic_info()
        if "routes" in self.basic_info.keys():
            return self.basic_info["routes"]
        return []

    def crawl_namespaces(self, ns):
        """Crawls all accessible get routes defined for the specified namespace."""
        namespaces = self.get_namespaces()
        routes = self.get_routes()
        ns_data = {}
        if ns != "all" and ns not in namespaces:
            raise NSNotFoundException
        for url, route in routes.items():
            if "namespace" not in route.keys() or "endpoints" not in route.keys():
                continue
            url_as_ns = url.lstrip("/")
            if "(?P<" in url or url_as_ns in namespaces:
                continue
            if (
                ns != "all"
                and route["namespace"] != ns
                or route["namespace"] in ["wp/v2", ""]
            ):
                continue
            for endpoint in route["endpoints"]:
                if "GET" not in endpoint["methods"]:
                    continue
                keep = True
                if len(endpoint["args"]) > 0 and type(endpoint["args"]) is dict:
                    for name, arg in endpoint["args"].items():  # noqa: B007
                        if arg["required"]:
                            keep = False
                if keep:
                    rest_url = url_path_join(self.url, self.api_path, url)
                    try:
                        ns_request = self.s.get(rest_url)
                        ns_data[url] = get_content_as_json(ns_request)
                    except Exception:
                        continue
        return ns_data

    def get_obj_by_id_helper(self, cache, obj_id, url, use_cache=True) -> list[dict]:
        """Retrieve an object from the cache or get it if not present.

        Args:
            cache: cache object for this type
            obj_id: id of the object to fetch
            url: URL formatting template containing "%d" where the ID should be substituted
            use_cache: whether to use the cache or force a re-fetch

        Returns:
            A list containing the returned object, empty if not retrievable.
        """
        if use_cache and cache is not None:
            obj = get_by_id(cache, obj_id)
            if obj is not None:
                return [obj]
        obj = self.crawl_single_page(url % obj_id)
        if type(obj) is dict:
            return [obj]
        return []

    def get_obj_by_id(self, obj_type, obj_id, use_cache=True):
        """Returns a list of maximum one object specified by its type and ID.

        Also returns an empty list if the ID does not exist.

        Args:
            obj_type: the type of the object (ex. POST)
            obj_id: the ID of the object to fetch
            use_cache: if the cache should be used to avoid useless
                requests
        """
        if obj_type == WPApi.USER:
            return self.get_obj_by_id_helper(
                self.users, obj_id, "wp/v2/users/%d", use_cache
            )
        if obj_type == WPApi.TAG:
            return self.get_obj_by_id_helper(
                self.tags, obj_id, "wp/v2/tags/%d", use_cache
            )
        if obj_type == WPApi.CATEGORY:
            return self.get_obj_by_id_helper(
                self.categories, obj_id, "wp/v2/categories/%d", use_cache
            )
        if obj_type == WPApi.POST:
            return self.get_obj_by_id_helper(
                self.posts, obj_id, "wp/v2/posts/%d", use_cache
            )
        if obj_type == WPApi.PAGE:
            return self.get_obj_by_id_helper(
                self.pages, obj_id, "wp/v2/pages/%d", use_cache
            )
        if obj_type == WPApi.COMMENT:
            return self.get_obj_by_id_helper(
                self.comments, obj_id, "wp/v2/comments/%d", use_cache
            )
        if obj_type == WPApi.MEDIA:
            return self.get_obj_by_id_helper(
                self.comments, obj_id, "wp/v2/media/%d", use_cache
            )
        return []

    def get_obj_list(self, obj_type, start, limit, cache, kwargs=None):
        """Returns a list of maximum limit objects specified by the starting object offset.

        Args:
            obj_type: the type of the object (ex. POST)
            start: the offset of the first object to return
            limit: the maximum number of objects to return
            cache: if the cache should be used to avoid useless requests
            kwargs: additional parameters to pass to the function (for
                POST only)
        """
        kwargs = kwargs or {}

        get_func = None
        if obj_type == WPApi.USER:
            get_func = self.get_users
        elif obj_type == WPApi.TAG:
            get_func = self.get_tags
        elif obj_type == WPApi.CATEGORY:
            get_func = self.get_categories
        elif obj_type == WPApi.PAGE:
            get_func = self.get_pages
        elif obj_type == WPApi.COMMENT:
            get_func = self.get_comments
        elif obj_type == WPApi.MEDIA:
            get_func = self.get_media
        elif obj_type == WPApi.NAMESPACE:
            get_func = self.get_namespaces

        if get_func is not None:
            return get_func(start=start, num=limit, force=not cache)
        elif obj_type == WPApi.POST:
            return self.get_posts(start=start, num=limit, force=not cache, **kwargs)
        return []

    def search(self, obj_types, keywords, start, limit):
        """Looks for data with the specified keywords of the given types.

        Args:
            obj_types: a list of the desired object types to look for
            keywords: the keywords to look for
            start: a start index
            limit: the max number to return

        Returns:
            a dict of lists of objects sorted by types
        """
        out = {}
        if WPApi.ALL_TYPES in obj_types or len(obj_types) == 0:
            obj_types = [
                WPApi.POST,
                WPApi.CATEGORY,
                WPApi.TAG,
                WPApi.PAGE,
                WPApi.COMMENT,
                WPApi.MEDIA,
                WPApi.USER,
            ]  # All supported types for search
        for t in obj_types:
            if t == WPApi.POST:
                out[t] = self.crawl_pages(
                    "wp/v2/posts?page=%d",
                    start=start,
                    num=limit,
                    search_terms=keywords,
                    display_progress=False,
                )[0]
            elif t == WPApi.CATEGORY:
                out[t] = self.crawl_pages(
                    "wp/v2/categories?page=%d",
                    start=start,
                    num=limit,
                    search_terms=keywords,
                    display_progress=False,
                )[0]
            elif t == WPApi.TAG:
                out[t] = self.crawl_pages(
                    "wp/v2/tags?page=%d",
                    start=start,
                    num=limit,
                    search_terms=keywords,
                    display_progress=False,
                )[0]
            elif t == WPApi.PAGE:
                out[t] = self.crawl_pages(
                    "wp/v2/pages?page=%d",
                    start=start,
                    num=limit,
                    search_terms=keywords,
                    display_progress=False,
                )[0]
            elif t == WPApi.COMMENT:
                out[t] = self.crawl_pages(
                    "wp/v2/comments?page=%d",
                    start=start,
                    num=limit,
                    search_terms=keywords,
                    display_progress=False,
                )[0]
            elif t == WPApi.MEDIA:
                out[t] = self.crawl_pages(
                    "wp/v2/media?page=%d",
                    start=start,
                    num=limit,
                    search_terms=keywords,
                    display_progress=False,
                )[0]
            elif t == WPApi.USER:
                out[t] = self.crawl_pages(
                    "wp/v2/users?page=%d",
                    start=start,
                    num=limit,
                    search_terms=keywords,
                    display_progress=False,
                )[0]
        return out



import dataclasses
import json
import logging
from pathlib import Path
from typing import Any, Optional

import numpy as np
import pandas as pd
from langcodes import Language
from pandas import DataFrame
from pandas import Timestamp as PdTimestamp


def load_from_path(path: Path) -> Optional[dict]:
    """Loads and parses a JSON file.

    Args:
        path: The path to load

    Returns:
        The decoded JSON object. None if the file does not exist.
    """
    if not path.is_file():
        return None

    with open(path) as f:
        return json.load(f)


def load_df(path: Path, index_col: str = "id") -> Optional[pd.DataFrame]:
    """Load a JSON file from a path into a dataframe and normalize.

    Will flatten the structure an infinite number of levels by
    concatenating child keys with ".".

    Args:
        path: The path to load
        index_col: The key from the JSON to use as the index

    Returns:
        A dataframe with the flattened JSON. None if the file does not exist or is an empty array.
    """
    data_raw = load_from_path(path)

    if data_raw is None:
        logging.info(f"File {path} does not exist, skipping this data type.")
        return None

    if len(data_raw) == 0:
        logging.info(f"File {path} has no data, skipping this data type.")
        return None

    return pd.json_normalize(data_raw).set_index(index_col)


def _set_nested_keys(row_dict: dict, split_key: list[str], val: Any):
    """Set a value in the dictionary with nested keys.

    Args:
        row_dict: The dictionary to set the value
        split_key: A list of 0 or more nested keys
        val: The value to set

    Raises:
        ValueError: If part of the nested key is already set to a value (for example,
            the dataframe has columns ``a`` and ``a.b``)

    Returns:
        The dictionary with the nested value set
    """
    current = row_dict
    for i in range(len(split_key)):
        key = split_key[i]
        if key not in current:
            if i == len(split_key) - 1:
                current[key] = val
            else:
                current[key] = {}
        else:
            if type(current[key]) is not dict:
                raise ValueError(f"Dictionary key {current[key]} is already set.")

        current = current[key]

    return row_dict


def df_denormalize_to_dict(df: pd.DataFrame, sep: str = "."):
    """Unflatten a dataframe with dot notation keys into a nested dictionary.

    Opposite of ``pd.json_normalize``.

    Args:
        df: the dataframe
        sep: the key separator

    Returns:
        A list of dictionaries with unflattened keys.
    """
    result = []
    for _, row in df.iterrows():
        parsed_row = {}
        for idx, val in row.items():
            keys = str(idx).split(sep)
            parsed_row = _set_nested_keys(parsed_row, keys, val)

        result.append(parsed_row)
    return result


def _remove_nan(value: Any) -> Any:
    """Replaces NA and NaN values with None.

    Supports replacing the values of lists and dictionaries.

    Args:
        value: The value to check

    Returns:
        The value with NA and NaN replaced.
    """
    if isinstance(value, dict):
        for dkey, dvalue in value.items():
            value[dkey] = _remove_nan(dvalue)
    elif isinstance(value, list):
        for i, item in enumerate(value):
            value[i] = _remove_nan(item)
    elif isinstance(value, float) and np.isnan(value):
        value = None
    elif value is pd.NA:
        value = None

    return value


class JSONEncoder(json.JSONEncoder):
    """Custom JSONEncoder to serialise site data."""

    def default(self, o: Any) -> Any:
        """Serialise custom objects.

        .. list-table
            :header-rows: 1
            * Type, Strategy
            * - ``pd.Timestamp``, ``isoformat()``
            * - Dataclasses, ``dataclasses.asdict()``
        """
        if type(o) == PdTimestamp:
            return o.isoformat()
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if type(o) == Language:
            return str(o)

        return super().default(o)


def _export_file(raw: Any, path: Path):
    with open(path, "w") as f:
        json.dump(raw, f, cls=JSONEncoder, indent=2, allow_nan=False)


def export_df(df: Optional[DataFrame], path: Path) -> None:
    """Export a dataframe to the path as JSON.

    Nests dot-notation columns into nested dictionaries.

    Args:
        df: the dataframe
        path: the destination path
    """
    if df is None:
        _export_file([], path)
    else:
        out_df = df.reset_index(names="id")
        denormalized = _remove_nan(df_denormalize_to_dict(out_df))
        _export_file(denormalized, path)



import logging
from pathlib import Path
from typing import Optional

import pandas as pd
from pandas import DataFrame
from tqdm.auto import tqdm

from wpextract.extractors.data.images import resolve_images
from wpextract.extractors.data.link_resolver import resolve_links
from wpextract.extractors.data.links import LinkRegistry
from wpextract.extractors.io import load_df
from wpextract.parse.content import extract_content_data
from wpextract.parse.html import extract_html_text, parse_html
from wpextract.parse.translations import PickerListType, extract_translations
from wpextract.parse.translations._resolver import TranslationLink
from wpextract.scrape.scrape import load_scrape
from wpextract.util.locale import extract_locale

EXPORT_COLUMNS = [
    "author",
    "categories",
    "comment_status",
    "content.rendered",
    "content.text",
    "date_gmt",
    "embeds",
    "excerpt.rendered",
    "excerpt.text",
    "featured_media",
    "images",
    "language",
    "link",
    "link_locale",
    "links.external",
    "links.internal",
    "modified_gmt",
    "og_image_url",
    "slug",
    "status",
    "sticky",
    "tags",
    "title.rendered",
    "title.text",
    "translations",
    "yoast_head_json.title",
]

RENAME_COLUMNS = {
    "title.rendered": "title.html",
    "content.rendered": "content.html",
    "excerpt.rendered": "excerpt.html",
    "yoast_head_json.title": "page_title",
}


def load_posts(
    path: Path,
    link_registry: LinkRegistry,
    scrape_urls_files: dict[str, Path],
    translation_pickers: Optional[PickerListType] = None,
) -> Optional[pd.DataFrame]:
    """Load the posts from a JSON file.

    The JSON file is expected to be in the response format of the WordPress posts API.

    Args:
        path: The path to the JSON file
        link_registry: The Link Registry to populate
        scrape_urls_files: A dictionary of site URLs to scrape file paths
        translation_pickers: Custom list of translation pickers.

    Returns:
        A dataframe of the posts.
    """
    posts_df = load_df(path)

    if posts_df is None:
        return None

    # Convert times
    posts_df["date_gmt"] = pd.to_datetime(posts_df["date_gmt"])
    posts_df["modified_gmt"] = pd.to_datetime(posts_df["modified_gmt"])
    posts_df = posts_df.drop(["date", "modified"], axis=1)

    # yoast_head_json.og_image is a list containing 0 or 1 image dictionaries
    # Get the "url" property if there is an image
    if "yoast_head_json.title" in posts_df.columns:
        posts_df["og_image_url"] = posts_df["yoast_head_json.og_image"].apply(
            lambda image: image[0]["url"]
            if not isinstance(image, float) and len(image) > 0
            else None
        )
    else:
        posts_df["og_image_url"] = None

    posts_df["link_locale"] = posts_df["link"].apply(extract_locale)

    posts_df["title.text"] = posts_df["title.rendered"].apply(extract_html_text)
    posts_df["excerpt.text"] = posts_df["excerpt.rendered"].apply(extract_html_text)

    tqdm.pandas(desc="Parsing Content")
    posts_df["content.bs"] = posts_df["content.rendered"].progress_apply(parse_html)

    if scrape_urls_files != {}:
        tqdm.pandas(desc="Parsing Scrape")
        posts_df["scrape_bs"] = posts_df["link"].progress_apply(
            lambda link: load_scrape(scrape_urls_files, link)
        )
        posts_df[["language", "translations"]] = posts_df.apply(
            lambda r: extract_translations(
                r["scrape_bs"], r["link"], translation_pickers
            ),
            axis=1,
        )
    else:
        logging.info("SKipping translation extraction")

    link_registry.add_linkables(
        "post", posts_df["link"].to_list(), posts_df.index.to_list()
    )

    tqdm.pandas(desc="Extracting from text")
    posts_df[
        ["content.text", "links.internal", "links.external", "embeds", "images"]
    ] = posts_df.progress_apply(
        lambda r: extract_content_data(r["content.bs"], r["link"]), axis=1
    )

    posts_df = posts_df[posts_df.columns.intersection(EXPORT_COLUMNS)]
    posts_df = posts_df.rename(columns=RENAME_COLUMNS, errors="ignore")

    return posts_df


def resolve_post_links(registry: LinkRegistry, posts_df: DataFrame) -> DataFrame:
    """Look up the internal links of each post in the registry.

    Args:
        registry: A filled link registry
        posts_df: The processed posts dataframe

    Returns:
        The posts dataframe with link data resolved.
    """
    posts_df["links.internal"] = posts_df["links.internal"].apply(
        lambda links: resolve_links(registry, links)
    )

    return posts_df


def resolve_post_media(registry: LinkRegistry, posts_df: DataFrame) -> DataFrame:
    """Look up the images of each post.

    Args:
        registry: A filled link registry
        posts_df: The processed posts dataframe

    Returns:
        The posts dataframe with link data resolved.
    """
    posts_df["images"] = posts_df["images"].apply(
        lambda media: resolve_images(registry, media)
    )

    return posts_df


def ensure_translations_undirected(posts_df: DataFrame) -> DataFrame:
    """Create translation relationships if they are not bidirectional.

    For example, if A -> B, but B -> A does not exist, then create B -> A.

    Translations will need to be resolved again after this step.

    Args:
        posts_df: A processed posts dataframe.

    Returns:
        The posts dataframe with bidirectional translations.
    """
    new_translations = []
    for post in posts_df.itertuples():
        for translation_obj in post.translations:
            if translation_obj.destination is None:
                logging.debug(
                    f"Unable to verify translation for {post.Index}, unresolved."
                )
                continue

            translation_id = translation_obj.destination.idx
            translation_post = posts_df.loc[translation_id]

            if not any(
                [
                    translation.destination is not None
                    and translation.destination.idx == post.Index
                    for translation in translation_post.translations
                ]
            ):
                logging.debug(
                    f"{post.Index} -> {translation_obj.destination.idx} existed "
                    "but reverse did not"
                )
                new_translations.append(
                    (
                        translation_id,
                        TranslationLink(
                            text=None,
                            href=post.link,
                            lang=post.language,
                            destination=None,
                        ),
                    )
                )

    for new_t_id, new_t_link in new_translations:
        posts_df.loc[new_t_id, "translations"].append(new_t_link)

    return posts_df


def resolve_post_translations(registry: LinkRegistry, posts_df: DataFrame) -> DataFrame:
    """Look up the translation links of each post in the registry.

    Args:
        registry: A filled link registry
        posts_df: The processed posts dataframe

    Returns:
        The posts dataframe with link data resolved
    """
    posts_df["translations"] = posts_df["translations"].apply(
        lambda ts: resolve_links(registry, ts)
    )

    return posts_df



import json
import logging
from pathlib import Path

from bs4 import BeautifulSoup
from tqdm.auto import tqdm

from wpextract.scrape.processor import extract_self_url, self_url_strainer

# Increment to invalidate old caches
SCRAPE_CRAWL_VERSION = 1


class ScrapeCrawl:
    """Crawl a scraped website to get original page URLs.

    Finds the original URL of every HTML file in the crawl from either the canonical
    link tag or the OpenGraph URL meta tag.

    Crawl results are stored in ``url_cache.json`` in the provided ``root_path``.
    If this file is present, the scrape will not be re-crawled. It contains the
    ``found_pages`` and ``failed_docs`` properties.

    Attributes:
        found_pages: Dictionary of path relative to `root_path` and the discovered URL.
        failed_docs: List of paths relative to `root_path` where a URL could
            not be discovered.
    """

    found_pages: dict[str, str]
    failed_docs: list[str]
    crawled = False

    def __init__(self, root_path: Path):
        """Init a new crawl instance.

        Args:
            root_path: The root path of the crawl (i.e. ``/`` on the website)

        Raises:
            ValueError: If the root path is not a directory.
        """
        if not root_path.is_dir():
            raise ValueError(f"Root path {root_path} is not a directory")

        self.root_path = root_path
        self.found_pages = {}
        self.failed_docs = []

        if self._get_cache_path().is_file():
            self._import()

    def _get_cache_path(self) -> Path:
        return self.root_path / "url_cache.json"

    def _export(self):
        with open(self._get_cache_path(), "w") as f:
            json.dump(
                {
                    "found": self.found_pages,
                    "failed": self.failed_docs,
                    "version": SCRAPE_CRAWL_VERSION,
                },
                f,
            )

    def _import(self):
        with open(self._get_cache_path()) as f:
            data = json.load(f)

        if "version" not in data or data["version"] != SCRAPE_CRAWL_VERSION:
            logging.info("Scrape crawl cache is out of date, re-crawling.")
            return

        self.found_pages = data["found"]
        self.failed_docs = data["failed"]
        self.crawled = True

    def crawl(self) -> None:
        """Perform the crawl or use the cached result."""
        if self.crawled:
            logging.info("Skipping scrape crawl, cached results available.")
            return

        files = list(self.root_path.glob("**/*.html"))

        for path in tqdm(files, desc="Crawling Scrape"):
            relative_path = str(path.relative_to(self.root_path))

            doc = BeautifulSoup(
                path.read_text(errors="ignore"),
                "lxml",
                parse_only=self_url_strainer,
            )
            doc_url = extract_self_url(doc)

            if doc_url is None:
                self.failed_docs.append(path)
                logging.warning(f'Failed to find self-URL in doc "{path}"')
                continue

            if doc_url in self.found_pages:
                logging.info(
                    f"URL {doc_url} retrieved for {relative_path}, but has already been"
                    f"found for {self.found_pages[relative_path]}"
                )
                continue

            self.found_pages[relative_path] = doc_url

        self._export()

    def get_link_abs_path(self) -> dict[str, Path]:
        """Get the mapping as original link to absolute file path.

        Absolute paths are resolved with the ``root_path`` and relative path from the
        crawl. Webpages may not exist if the crawl was loaded from cache.

        Returns:
            Dictionary of URLs to absolute paths.
        """
        return {
            link: (self.root_path / Path(relative_path)).resolve()
            for relative_path, link in self.found_pages.items()
        }



class NoWordpressApi(Exception):
    """No API is available at the given URL."""

    pass


class WordPressApiNotV2(Exception):
    """The WordPress V2 API is not available."""

    pass


class NSNotFoundException(Exception):
    """The specified namespace does not exist."""

    pass
